package ru.dbotthepony.mc.otm.capability;

import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.food.FoodData;
import net.minecraft.world.item.ItemStack;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.capabilities.ICapabilityProvider;
import net.minecraftforge.common.util.INBTSerializable;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.energy.IEnergyStorage;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.math.BigDecimal;

public class AndroidCapability implements ICapabilityProvider, IAndroidCapability, INBTSerializable<CompoundTag> {
	private final LivingEntity ent;
	private BigDecimal energy_stored = new BigDecimal(0);
	private BigDecimal energy_stored_max = new BigDecimal(32000);
	private ItemStack battery = ItemStack.EMPTY;

	public static final BigDecimal ENERGY_FOR_HUNGER_POINT = new BigDecimal(200);

	public AndroidCapability(LivingEntity ent) {
		this.ent = ent;
	}

	@Override
	@Nonnull
	public CompoundTag serializeNBT() {
		CompoundTag tag = new CompoundTag();
		tag.putString("energy_stored", energy_stored.toString());
		tag.putString("energy_stored_max", energy_stored_max.toString());
		tag.put("battery", battery.serializeNBT());
		return tag;
	}

	@Override
	public void deserializeNBT(CompoundTag compound) {
		if (compound.contains("energy_stored"))
			energy_stored = new BigDecimal(compound.getString("energy_stored"));

		if (compound.contains("energy_stored_max"))
			energy_stored_max = new BigDecimal(compound.getString("energy_stored_max"));

		if (compound.contains("battery"))
			battery = ItemStack.of(compound.getCompound("battery"));
	}

	@Override
	public void tick() {
		if (ent instanceof ServerPlayer) {
			FoodData stats = ((ServerPlayer) ent).getFoodData();

			if (stats.needsFood()) {
				int food_points_missing = 20 - stats.getFoodLevel();
				boolean food_changed = false;
				float saturation_points_missing = 20F - stats.getSaturationLevel();
				boolean saturation_changed = false;

				while (food_points_missing > 0) {
					if (this.extractEnergyInner(ENERGY_FOR_HUNGER_POINT, true).equals(ENERGY_FOR_HUNGER_POINT)) {
						this.extractEnergyInner(ENERGY_FOR_HUNGER_POINT, false);
						food_points_missing--;
						food_changed = true;
					} else {
						break;
					}
				}

				while (saturation_points_missing >= 1F) {
					if (this.extractEnergyInner(ENERGY_FOR_HUNGER_POINT, true).equals(ENERGY_FOR_HUNGER_POINT)) {
						this.extractEnergyInner(ENERGY_FOR_HUNGER_POINT, false);
						saturation_points_missing--;
						saturation_changed = true;
					} else {
						break;
					}
				}

				if (food_changed) {
					stats.setFoodLevel(20 - food_points_missing);
				}

				if (saturation_changed) {
					stats.setSaturation(20F - saturation_points_missing);
				}

				OverdriveThatMatters.LOGGER.info("New power level {}", getBatteryLevel());
			}
		}
	}

	@Nonnull
	@Override
	public BigDecimal extractEnergyOuter(BigDecimal howMuch, boolean simulate) {
		return BigDecimal.ZERO;
	}

	@Nonnull
	@Override
	public BigDecimal extractEnergyInner(BigDecimal howMuch, boolean simulate) {
		if (battery != ItemStack.EMPTY) {
			LazyOptional<IMatteryEnergyStorage> get_mattery = battery.getCapability(MatteryCapability.ENERGY);

			if (get_mattery.isPresent()) {
				return get_mattery.resolve().get().extractEnergyInner(howMuch, simulate);
			}

			if (howMuch.compareTo(BigDecimal.ONE) >= 0) {
				LazyOptional<IEnergyStorage> get_energy = battery.getCapability(CapabilityEnergy.ENERGY);

				if (get_energy.isPresent()) {
					int converted = howMuch.compareTo(MatteryCapability.INT_MAX_VALUE) <= 0 ? howMuch.intValue() : Integer.MAX_VALUE;
					return new BigDecimal(get_energy.resolve().get().extractEnergy(converted, simulate));
				}
			}
		}

		BigDecimal new_energy = energy_stored.subtract(howMuch, MatteryCapability.ROUND_RULES).max(BigDecimal.ZERO);
		BigDecimal diff = energy_stored.subtract(new_energy, MatteryCapability.ROUND_RULES);

		if (!simulate) {
			energy_stored = new_energy;
		}

		return diff;
	}

	@Nonnull
	@Override
	public BigDecimal receiveEnergyOuter(BigDecimal howMuch, boolean simulate) {
		if (battery != ItemStack.EMPTY) {
			LazyOptional<IMatteryEnergyStorage> get_mattery = battery.getCapability(MatteryCapability.ENERGY);

			if (get_mattery.isPresent()) {
				return get_mattery.resolve().get().receiveEnergyOuter(howMuch, simulate);
			}

			if (howMuch.compareTo(BigDecimal.ONE) >= 0) {
				LazyOptional<IEnergyStorage> get_energy = battery.getCapability(CapabilityEnergy.ENERGY);

				if (get_energy.isPresent()) {
					int converted = howMuch.compareTo(MatteryCapability.INT_MAX_VALUE) <= 0 ? howMuch.intValue() : Integer.MAX_VALUE;
					return new BigDecimal(get_energy.resolve().get().receiveEnergy(converted, simulate));
				}
			}
		}

		BigDecimal new_energy = energy_stored.add(howMuch, MatteryCapability.ROUND_RULES).min(energy_stored_max);
		BigDecimal diff = new_energy.subtract(energy_stored, MatteryCapability.ROUND_RULES);

		if (!simulate) {
			energy_stored = new_energy;
		}

		return diff;
	}

	@Nonnull
	@Override
	public BigDecimal receiveEnergyInner(BigDecimal howMuch, boolean simulate) {
		return receiveEnergyOuter(howMuch, simulate);
	}

	@Override
	public LivingEntity getEntity() {
		return this.ent;
	}

	@Nonnull
	@Override
	public ItemStack getBatteryItemStack() {
		return battery;
	}

	@Override
	public void setBatteryItemStack(@Nonnull ItemStack stack) {
		battery = stack;
	}

	@Override
	@Nonnull
	public BigDecimal getBatteryLevel() {
		if (battery != ItemStack.EMPTY) {
			LazyOptional<IMatteryEnergyStorage> get_mattery = battery.getCapability(MatteryCapability.ENERGY);

			if (get_mattery.isPresent()) {
				return get_mattery.resolve().get().getBatteryLevel();
			}

			LazyOptional<IEnergyStorage> get_energy = battery.getCapability(CapabilityEnergy.ENERGY);

			if (get_energy.isPresent()) {
				return new BigDecimal(get_energy.resolve().get().getEnergyStored());
			}
		}

		return energy_stored;
	}

	@Override
	@Nonnull
	public BigDecimal getMaxBatteryLevel() {
		if (battery != ItemStack.EMPTY) {
			LazyOptional<IMatteryEnergyStorage> get_mattery = battery.getCapability(MatteryCapability.ENERGY);

			if (get_mattery.isPresent()) {
				return get_mattery.resolve().get().getMaxBatteryLevel();
			}

			LazyOptional<IEnergyStorage> get_energy = battery.getCapability(CapabilityEnergy.ENERGY);

			if (get_energy.isPresent()) {
				return new BigDecimal(get_energy.resolve().get().getMaxEnergyStored());
			}
		}

		return energy_stored_max;
	}

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (cap == MatteryCapability.ANDROID) {
			return LazyOptional.of(() -> this).cast();
		}

		return LazyOptional.empty();
	}
}
